@@ -1,13 +1,23 @@
 import io
 import logging
 import os
+import threading
 import warnings
-from hashlib import sha256
+from distutils.version import LooseVersion
+from errno import ESPIPE
 from glob import has_magic
+from hashlib import sha256
 
+from .config import apply_config
 from .dircache import DirCache
 from .transaction import Transaction
-from .utils import read_block, tokenize, stringify_path, other_paths
+from .utils import (
+    get_package_version_without_import,
+    other_paths,
+    read_block,
+    stringify_path,
+    tokenize,
+)
 
 logger = logging.getLogger("fsspec")
 
@@ -39,13 +49,20 @@ class _Cached(type):
         # collecting instances when all other references are gone. To really
         # delete a FileSystem, the cache must be cleared.
         cls._cache = {}
+        cls._pid = os.getpid()
 
     def __call__(cls, *args, **kwargs):
+        kwargs = apply_config(cls, kwargs)
         extra_tokens = tuple(
             getattr(cls, attr, None) for attr in cls._extra_tokenize_attributes
         )
-        token = tokenize(cls, *args, *extra_tokens, **kwargs)
+        token = tokenize(
+            cls, cls._pid, threading.get_ident(), *args, *extra_tokens, **kwargs
+        )
         skip = kwargs.pop("skip_instance_cache", False)
+        if os.getpid() != cls._pid:
+            cls._cache.clear()
+            cls._pid = os.getpid()
         if not skip and cls.cachable and token in cls._cache:
             return cls._cache[token]
         else:
@@ -64,12 +81,16 @@ class _Cached(type):
             return obj
 
 
-try:  # optionally derive from pyarrow's FileSystem, if available
-    # TODO: it should be possible to disable this
-    import pyarrow as pa
+pa_version = get_package_version_without_import("pyarrow")
+if pa_version and LooseVersion(pa_version) < LooseVersion("2.0"):
+    try:
+        import pyarrow as pa
 
-    up = pa.filesystem.DaskFileSystem
-except ImportError:
+        up = pa.filesystem.DaskFileSystem
+    except ImportError:  # pragma: no cover
+        # pyarrow exists but doesn't import for some reason
+        up = object
+else:  # pragma: no cover
     up = object
 
 
@@ -122,6 +143,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         self._cached = True
         self._intrans = False
         self._transaction = None
+        self._invalidated_caches_in_transaction = []
         self.dircache = DirCache(**storage_options)
 
         if storage_options.pop("add_docs", None):
@@ -150,10 +172,12 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
     @classmethod
     def _strip_protocol(cls, path):
-        """ Turn path from fully-qualified to file-system-specific
+        """Turn path from fully-qualified to file-system-specific
 
         May require FS-specific handling, e.g., for relative paths or links.
         """
+        if isinstance(path, list):
+            return [cls._strip_protocol(p) for p in path]
         path = stringify_path(path)
         protos = (cls.protocol,) if isinstance(cls.protocol, str) else cls.protocol
         for protocol in protos:
@@ -180,7 +204,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
     @classmethod
     def current(cls):
-        """ Return the most recently created FileSystem
+        """Return the most recently created FileSystem
 
         If no instance has been created, then create one with defaults
         """
@@ -210,6 +234,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """Finish write transaction, non-context version"""
         self.transaction.complete()
         self._transaction = None
+        # The invalid cache must be cleared after the transcation is completed.
+        for path in self._invalidated_caches_in_transaction:
+            self.invalidate_cache(path)
+        self._invalidated_caches_in_transaction.clear()
 
     def invalidate_cache(self, path=None):
         """
@@ -221,7 +249,12 @@ class AbstractFileSystem(up, metaclass=_Cached):
             If None, clear all listings cached else listings at or under given
             path.
         """
-        pass  # not necessary to implement, may have no cache
+        # Not necessary to implement invalidation mechanism, may have no cache.
+        # But if have, you should call this method of parent class from your
+        # subclass to ensure expiring caches after transacations correctly.
+        # See the implementaion of FTPFileSystem in ftp.py
+        if self._intrans:
+            self._invalidated_caches_in_transaction.append(path)
 
     def mkdir(self, path, create_parents=True, **kwargs):
         """
@@ -309,12 +342,15 @@ class AbstractFileSystem(up, metaclass=_Cached):
         but contains nothing), None if not in cache.
         """
         parent = self._parent(path)
+        if path.rstrip("/") in self.dircache:
+            return self.dircache[path.rstrip("/")]
         try:
-            return self.dircache[path]
-        except KeyError:
-            pass
-        try:
-            files = [f for f in self.dircache[parent] if f["name"] == path]
+            files = [
+                f
+                for f in self.dircache[parent]
+                if f["name"] == path
+                or (f["name"] == path.rstrip("/") and f["type"] == "directory")
+            ]
             if len(files) == 0:
                 # parent dir was listed but did not contain this file
                 raise FileNotFoundError(path)
@@ -323,7 +359,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
             pass
 
     def walk(self, path, maxdepth=None, **kwargs):
-        """ Return all files belows path
+        """Return all files belows path
 
         List all files, recursing into subdirectories; output is iterator-style,
         like ``os.walk()``. For a simple list of files, ``find()`` is available.
@@ -398,7 +434,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         path = self._strip_protocol(path)
         out = dict()
         detail = kwargs.pop("detail", False)
-        for path, dirs, files in self.walk(path, maxdepth, detail=True, **kwargs):
+        for _, dirs, files in self.walk(path, maxdepth, detail=True, **kwargs):
             if withdirs:
                 files.update(dirs)
             out.update({info["name"]: info for name, info in files.items()})
@@ -446,7 +482,11 @@ class AbstractFileSystem(up, metaclass=_Cached):
         the same as ``ls(path)``, returning only files.
 
         We support ``"**"``,
-        ``"?"`` and ``"[..]"``.
+        ``"?"`` and ``"[..]"``. We do not support ^ for pattern negation.
+
+        Search path names that contain embedded characters special to this
+        implementation of glob may not produce expected results;
+        e.g., 'foo/bar/*starredfilename*'.
 
         kwargs are passed to ``ls``.
         """
@@ -480,12 +520,16 @@ class AbstractFileSystem(up, metaclass=_Cached):
         elif "/" in path[:ind]:
             ind2 = path[:ind].rindex("/")
             root = path[: ind2 + 1]
-            depth = 20 if "**" in path else path[ind2 + 1 :].count("/") + 1
+            depth = None if "**" in path else path[ind2 + 1 :].count("/") + 1
         else:
             root = ""
-            depth = 20 if "**" in path else 1
+            depth = None if "**" in path else path[ind + 1 :].count("/") + 1
 
         allpaths = self.find(root, maxdepth=depth, withdirs=True, detail=True, **kwargs)
+        # Escape characters special to python regex, leaving our supported
+        # special characters in place.
+        # See https://www.gnu.org/software/bash/manual/html_node/Pattern-Matching.html
+        # for shell globbing details.
         pattern = (
             "^"
             + (
@@ -496,6 +540,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
                 .replace("(", r"\(")
                 .replace(")", r"\)")
                 .replace("|", r"\|")
+                .replace("^", r"\^")
+                .replace("$", r"\$")
+                .replace("{", r"\{")
+                .replace("}", r"\}")
                 .rstrip("/")
                 .replace("?", ".")
             )
@@ -514,10 +562,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
         else:
             return list(out)
 
-    def exists(self, path):
+    def exists(self, path, **kwargs):
         """Is there a file at the given path"""
         try:
-            self.info(path)
+            self.info(path, **kwargs)
             return True
         except:  # noqa: E722
             # any exception allowed bar FileNotFoundError?
@@ -588,9 +636,15 @@ class AbstractFileSystem(up, metaclass=_Cached):
         except:  # noqa: E722
             return False
 
-    def cat_file(self, path):
+    def cat_file(self, path, start=None, end=None, **kwargs):
         """ Get the content of a file """
-        return self.open(path, "rb").read()
+        # explicitly set buffering off?
+        with self.open(path, "rb", **kwargs) as f:
+            if start is not None:
+                f.seek(start)
+            if end is not None:
+                return f.read(end - f.tell())
+            return f.read()
 
     def pipe_file(self, path, value, **kwargs):
         """Set the bytes of given file"""
@@ -618,11 +672,18 @@ class AbstractFileSystem(up, metaclass=_Cached):
         else:
             raise ValueError("path must be str or dict")
 
-    def cat(self, path, recursive=False, **kwargs):
+    def cat(self, path, recursive=False, on_error="raise", **kwargs):
         """Fetch (potentially multiple) paths' contents
 
         Returns a dict of {path: contents} if there are multiple paths
         or the path has been otherwise expanded
+
+        on_error : "raise", "omit", "return"
+            If raise, an underlying exception will be raised (converted to KeyError
+            if the type is in self.missing_exceptions); if omit, keys with exception
+            will simply not be included in the output; if "return", all keys are
+            included in the output, but the value will be bytes or an exception
+            instance.
         """
         paths = self.expand_path(path, recursive=recursive)
         if (
@@ -630,7 +691,16 @@ class AbstractFileSystem(up, metaclass=_Cached):
             or isinstance(path, list)
             or paths[0] != self._strip_protocol(path)
         ):
-            return {path: self.cat_file(path, **kwargs) for path in paths}
+            out = {}
+            for path in paths:
+                try:
+                    out[path] = self.cat_file(path, **kwargs)
+                except Exception as e:
+                    if on_error == "raise":
+                        raise
+                    if on_error == "return":
+                        out[path] = e
+            return out
         else:
             return self.cat_file(paths[0])
 
@@ -687,7 +757,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
 
         Calls put_file for each source.
         """
-        from .implementations.local import make_path_posix, LocalFileSystem
+        from .implementations.local import LocalFileSystem, make_path_posix
 
         rpath = (
             self._strip_protocol(rpath)
@@ -717,18 +787,36 @@ class AbstractFileSystem(up, metaclass=_Cached):
     def cp_file(self, path1, path2, **kwargs):
         raise NotImplementedError
 
-    def copy(self, path1, path2, recursive=False, **kwargs):
-        """ Copy within two locations in the filesystem"""
+    def copy(self, path1, path2, recursive=False, on_error=None, **kwargs):
+        """Copy within two locations in the filesystem
+
+        on_error : "raise", "ignore"
+            If raise, any not-found exceptions will be raised; if ignore any
+            not-found exceptions will cause the path to be skipped; defaults to
+            raise unless recursive is true, where the default is ignore
+        """
+        if on_error is None and recursive:
+            on_error = "ignore"
+        elif on_error is None:
+            on_error = "raise"
+
         paths = self.expand_path(path1, recursive=recursive)
         path2 = other_paths(paths, path2)
         for p1, p2 in zip(paths, path2):
-            self.cp_file(p1, p2, **kwargs)
+            try:
+                self.cp_file(p1, p2, **kwargs)
+            except FileNotFoundError:
+                if on_error == "raise":
+                    raise
 
     def expand_path(self, path, recursive=False, maxdepth=None):
-        """Turn one or more globs or directories into a list of all matching files"""
+        """Turn one or more globs or directories into a list of all matching paths
+        to files or directories."""
         if isinstance(path, str):
             out = self.expand_path([path], recursive, maxdepth)
         else:
+            # reduce depth on each recursion level unless None or 0
+            maxdepth = maxdepth if not maxdepth else maxdepth - 1
             out = set()
             path = [self._strip_protocol(p) for p in path]
             for p in path:
@@ -736,10 +824,14 @@ class AbstractFileSystem(up, metaclass=_Cached):
                     bit = set(self.glob(p))
                     out |= bit
                     if recursive:
-                        out += self.expand_path(p)
+                        out |= set(
+                            self.expand_path(
+                                list(bit), recursive=recursive, maxdepth=maxdepth
+                            )
+                        )
                     continue
                 elif recursive:
-                    rec = set(self.find(p, withdirs=True))
+                    rec = set(self.find(p, maxdepth=maxdepth, withdirs=True))
                     out |= rec
                 if p not in out and (recursive is False or self.exists(p)):
                     # should only check once, for the root
@@ -785,7 +877,8 @@ class AbstractFileSystem(up, metaclass=_Cached):
     def _parent(cls, path):
         path = cls._strip_protocol(path.rstrip("/"))
         if "/" in path:
-            return cls.root_marker + path.rsplit("/", 1)[0]
+            parent = path.rsplit("/", 1)[0].lstrip(cls.root_marker)
+            return cls.root_marker + parent
         else:
             return cls.root_marker
 
@@ -796,7 +889,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         block_size=None,
         autocommit=True,
         cache_options=None,
-        **kwargs
+        **kwargs,
     ):
         """Return raw bytes-mode file-like from the file-system"""
         return AbstractBufferedFile(
@@ -806,7 +899,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
             block_size,
             autocommit,
             cache_options=cache_options,
-            **kwargs
+            **kwargs,
         )
 
     def open(self, path, mode="rb", block_size=None, cache_options=None, **kwargs):
@@ -850,14 +943,14 @@ class AbstractFileSystem(up, metaclass=_Cached):
                 block_size=block_size,
                 autocommit=ac,
                 cache_options=cache_options,
-                **kwargs
+                **kwargs,
             )
-            if not ac:
+            if not ac and "r" not in mode:
                 self.transaction.files.append(f)
             return f
 
     def touch(self, path, truncate=True, **kwargs):
-        """ Create empty file, or update timestamp
+        """Create empty file, or update timestamp
 
         Parameters
         ----------
@@ -878,7 +971,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         return sha256(str(self.info(path)).encode()).hexdigest()
 
     def read_block(self, fn, offset, length, delimiter=None):
-        """ Read a block of bytes from
+        """Read a block of bytes from
 
         Starting at ``offset`` of the file, read ``length`` bytes.  If
         ``delimiter`` is set then we ensure that the read starts and stops at
@@ -945,7 +1038,7 @@ class AbstractFileSystem(up, metaclass=_Cached):
         return json.dumps(
             dict(
                 **{"cls": cls, "protocol": proto, "args": self.storage_args},
-                **self.storage_options
+                **self.storage_options,
             )
         )
 
@@ -964,9 +1057,10 @@ class AbstractFileSystem(up, metaclass=_Cached):
         -------
         file system instance, not necessarily of this particular class.
         """
-        from .registry import _import_class, get_filesystem_class
         import json
 
+        from .registry import _import_class, get_filesystem_class
+
         dic = json.loads(blob)
         protocol = dic.pop("protocol")
         try:
@@ -1086,6 +1180,13 @@ class AbstractFileSystem(up, metaclass=_Cached):
         """
         raise NotImplementedError("Sign is not implemented for this filesystem")
 
+    def _isfilestore(self):
+        # Originally inherited from pyarrow DaskFileSystem. Keeping this
+        # here for backwards compatibility as long as pyarrow uses its
+        # legacy ffspec-compatible filesystems and thus accepts fsspec
+        # filesystems as well
+        return False
+
 
 class AbstractBufferedFile(io.IOBase):
     """Convenient class to derive from to provide buffering
@@ -1107,7 +1208,7 @@ class AbstractBufferedFile(io.IOBase):
         autocommit=True,
         cache_type="readahead",
         cache_options=None,
-        **kwargs
+        **kwargs,
     ):
         """
         Template for files with buffered reading and writing
@@ -1213,7 +1314,7 @@ class AbstractBufferedFile(io.IOBase):
         return self.loc
 
     def seek(self, loc, whence=0):
-        """ Set current file location
+        """Set current file location
 
         Parameters
         ----------
@@ -1224,7 +1325,7 @@ class AbstractBufferedFile(io.IOBase):
         """
         loc = int(loc)
         if not self.mode == "rb":
-            raise OSError("Seek only available in read mode")
+            raise OSError(ESPIPE, "Seek only available in read mode")
         if whence == 0:
             nloc = loc
         elif whence == 1:
@@ -1294,14 +1395,18 @@ class AbstractBufferedFile(io.IOBase):
         if self.offset is None:
             # Initialize a multipart upload
             self.offset = 0
-            self._initiate_upload()
+            try:
+                self._initiate_upload()
+            except:  # noqa: E722
+                self.closed = True
+                raise
 
         if self._upload_chunk(final=force) is not False:
             self.offset += self.buffer.seek(0, 2)
             self.buffer = io.BytesIO()
 
     def _upload_chunk(self, final=False):
-        """ Write one part of a multi-block file upload
+        """Write one part of a multi-block file upload
 
         Parameters
         ==========
@@ -1348,8 +1453,9 @@ class AbstractBufferedFile(io.IOBase):
 
         https://docs.python.org/3/library/io.html#io.RawIOBase.readinto
         """
-        data = self.read(len(b))
-        memoryview(b).cast("B")[: len(data)] = data
+        out = memoryview(b).cast("B")
+        data = self.read(out.nbytes)
+        out[: len(data)] = data
         return len(data)
 
     def readuntil(self, char=b"\n", blocks=None):
@@ -1412,10 +1518,12 @@ class AbstractBufferedFile(io.IOBase):
         return self.readinto(b)
 
     def close(self):
-        """ Close file
+        """Close file
 
         Finalizes writes, discards cache
         """
+        if getattr(self, "_unclosable", False):
+            return
         if self.closed:
             return
         if self.mode == "rb":
@@ -1443,7 +1551,8 @@ class AbstractBufferedFile(io.IOBase):
         return self.mode in {"wb", "ab"} and not self.closed
 
     def __del__(self):
-        self.close()
+        if not self.closed:
+            self.close()
 
     def __str__(self):
         return "<File-like object %s, %s>" % (type(self.fs).__name__, self.path)
